/*
 * Message.c
 *
 * Player to player messages module.
 */

#include "stratego.h"

#ifdef SYSV
#include <signal.h>
#else
#include <sys/filio.h>
#endif
#include <varargs.h>

#define REMISE_CONFIRMED	0
#define REMISE_DENIED		1

#define T_MOVE		'0'		/* Messages during play */
#define T_REMISE	'1'
#define T_REMISE_ANSWER	'2'

#define T_PLACED_TILE	'3'		/* Messages during setup */
#define T_SETUP_READY	'4'
#define T_CLEAR_BOARD	'5'

#define T_QUIT		'6'		/* General messages */
#define T_MAIL		'7'

#define MESSAGE_LENGTH	125

static char message_buffer[MESSAGE_LENGTH];
static char message_length[4];

static void receive_ready (), receive_move (), receive_placed_tile (),
	    receive_remise_answer (), receive_clear_board (),
	    receive_quit ();

void end_communications()
{
	if (players[1].fd > 0) {
		XtRemoveInput (EnemyID);
		players[1].fd = -1;
	}
}

static void read_data (fd, output, length)
int fd, length;
char *output;
{
	extern int errno;

	if (read (fd, output, length) == length)
		return;

	if (errno != 0)
		perror ("read_data()");

	printf("Communications breakdown. Host/Program crash on other side?\n");

	XtDestroyApplicationContext (app_con);
	exit (1);
}

XtInputCallbackProc EnemyInterrupt (client_data, source, id)
XtPointer client_data;
int *source;
XtInputId *id;
{
	if (players[1].fd == -1)
		return;

	read_data (players[1].fd, message_length, sizeof(message_length));
	read_data (players[1].fd, message_buffer, atoi (message_length));

	switch (message_buffer[0]) {
	case T_CLEAR_BOARD:
		receive_clear_board ();
		break;
	case T_SETUP_READY:
		receive_ready ();
		break;
	case T_MOVE:
		receive_move ();
		break;
	case T_QUIT:
		end_communications ();
		receive_quit ();
		return;
	case T_PLACED_TILE:
		receive_placed_tile ();
		break;
	case T_REMISE:
		RemiseOffered ();
		break;
	case T_REMISE_ANSWER:
		receive_remise_answer ();
		break;
	case T_MAIL:
		message ("%s: \"%s\"", players[1].name,
			 &message_buffer[2]);
		break;
	default:
		message ("Unknown input from opponent (type [%o] %c)",
			  message_buffer[0], message_buffer[0]);
	}
}

static void receive_clear_board ()
{
	int y, x;

	for (y = 0; y < 4; y++) {
		for (x = 0; x < MAX_COLS; x++) {
			if (board[y][x].value != FREE)
				AssignBoardPosition (&board[y][x],
						     &tiles[FREE], O_NOBODY);
		}
	}
}

static void receive_ready ()
{
	players[1].status = PL_SETUP_READY;
	if (players[0].status == PL_SETUP)
		message ("%s just finished his board setup. Hurry up %s!",
		 	 players[1].name, players[0].name);
	else
		start_play ();
}

static void receive_quit ()
{
	if (players[0].status <= PL_SETUP_READY)
		message ("%s decided to quit his setup!", players[1].name);
	else
		message ("You win %s, because %s admitted your superior strategy!", players[0].name, players[1].name);

	end_game ();
}

static void receive_placed_tile ()
{
	int x, y, tile_value;

	sscanf (&message_buffer[1], "%d %d %d", &x, &y, &tile_value);
	y = (MAX_ROWS - 1) - y;
	x = (MAX_COLS - 1) - x;

	AssignBoardPosition (&board[y][x], &tiles[tile_value], O_ENEMY);
}

static void receive_move ()
{
	int from_y, from_x, to_y, to_x;

	sscanf (&message_buffer[1], "%d %d %d %d", &from_y, &from_x,
				    		   &to_y, &to_x);

	from_x = (MAX_COLS - 1) - from_x;
	from_y = (MAX_ROWS - 1) - from_y;
	to_x   = (MAX_COLS - 1) - to_x;
	to_y   = (MAX_ROWS - 1) - to_y;

	switch (execute_move (&board[from_y][from_x],
			      &board[to_y][to_x])) {
	case REMISE_MOVE:
	case WIN_MOVE:
		end_game ();
		break;
	default:
		check_can_play (O_PLAYER);
		players[0].status = PL_PLAYING;

		EnableRemiseButton();
	}

}

static void receive_remise_answer ()
{
	static char msg_buf[100];
	int answer;

	sscanf (&message_buffer[1], "%d", &answer);
	
	if (answer == REMISE_DECLINED)
		sprintf (msg_buf, "%s declined your remise offer %s.",
			  players[1].name, players[0].name);

	else
		sprintf (msg_buf, "%s accepted your remise offer %s.",
			  players[1].name, players[0].name);

	(void) DialogBoxResult (msg_buf, NULL, 0, NO_CANCEL);

	if (answer == REMISE_ACCEPTED)
		end_game ();
}


/*
 * Send module.
 */

/*VARARGS0*/
static void send_buffer (va_alist)
va_dcl
{
	va_list arglist;
	char *fmt;

	if (players[0].status == PL_ENDGAME || players[0].setup_only)
		return;

	va_start (arglist);
	fmt = va_arg (arglist, char *);

	vsprintf (message_buffer, fmt, arglist);
	va_end (arglist);

	sprintf (message_length, "%d", strlen (message_buffer) + 1);

	write (players[1].fd, message_length, sizeof (message_length));
	write (players[1].fd, message_buffer, atoi (message_length));
}

void send_ready ()
{
	send_buffer ("%c", T_SETUP_READY);
}

void send_clear_board ()
{
	send_buffer ("%c", T_CLEAR_BOARD);
}

void send_tile_placed (board_pos)
BOARD_POS *board_pos;
{
	send_buffer ("%c %d %d %d", T_PLACED_TILE,
		     board_pos->x, board_pos->y, board_pos->value);
}

void send_move (from_pos, to_pos)
BOARD_POS *from_pos, *to_pos;
{
	send_buffer ("%c %d %d %d %d", T_MOVE,
		 from_pos->y, from_pos->x, to_pos->y, to_pos->x);
}

void send_remise_answer (answer)
int answer;
{
	send_buffer ("%c %d", T_REMISE_ANSWER, answer);
}

void send_remise_offer ()
{
	send_buffer ("%c", T_REMISE);
}

void send_quit ()
{
	send_buffer ("%c", T_QUIT);
	signal (SIGIO, SIG_IGN);
}

void send_mail (mail)
char *mail;
{
	if (players[0].status == PL_ENDGAME)
		return;

	send_buffer ("%c %s", T_MAIL, mail);
}
